from recordwhat.parsers.dbd_parsimonious import (
    dbd_grammar, dbdField, dbdRecordType, RecordWalker,
    stream_dbd, generate_all, stream_table, _BASE_FIELDS, _IMPORTS,
    prettytable_summary)
from recordwhat.parsers.generate import (load_dbd_derived, generate)
from recordwhat.util import read_file
from collections import OrderedDict
import pytest
from tempfile import NamedTemporaryFile, TemporaryFile
import os.path


@pytest.fixture
def test_rec():
    fields = OrderedDict()
    for pg, special in zip(('AA', 'BB'), ('', 'SPC_NOMOD')):
        for j in range(5):
            nm = '{}{}'.format(pg, j)
            fields[nm] = dbdField(
                nm,
                'DBF_CHAR',
                prompt='"test p {}"'.format(j),
                promptgroup='"{}"'.format(pg),
                special=special)
        fields['SEC'] = dbdField('SEC',
                                 'DBF_NOACCESS')

    return dbdRecordType(name='test',
                         fields=fields)


def test_round_trip(test_rec):
    rec = test_rec
    dbd_str = '\n'.join(stream_dbd(rec))

    with NamedTemporaryFile(suffix='.dbd', mode='wt',
                            encoding='utf-8') as f1, \
        NamedTemporaryFile(suffix='.dbd', mode='wt',
                           encoding='utf-8') as f2:
        f1.write(dbd_str)
        fn = os.path.basename(f1.name)
        f2.write('#somecomment\ninclude "aardvark.dbd"\n')
        f2.write('include "{}"'.format(fn))
        f1.flush()
        f2.flush()
        test_str = read_file(f2.name)

    p = dbd_grammar.parse(test_str)

    rec2 = RecordWalker().visit(p)

    assert rec == rec2['test']


_target_code = """\
from ophyd import (EpicsSignal, EpicsSignalRO)

from recordwhat import (RecordBase, _register_record_type,
                        FieldComponent as Cpt)


@_register_record_type('test')
class TestRecord(RecordBase):
    'Autogenerated class'

    # - "aa"
    test_p_0 = Cpt(EpicsSignal, '.AA0')
    test_p_1 = Cpt(EpicsSignal, '.AA1')
    test_p_2 = Cpt(EpicsSignal, '.AA2')
    test_p_3 = Cpt(EpicsSignal, '.AA3')
    test_p_4 = Cpt(EpicsSignal, '.AA4')

    # - "bb"
    test_p_0_bb0 = Cpt(EpicsSignalRO, '.BB0')
    test_p_1_bb1 = Cpt(EpicsSignalRO, '.BB1')
    test_p_2_bb2 = Cpt(EpicsSignalRO, '.BB2')
    test_p_3_bb3 = Cpt(EpicsSignalRO, '.BB3')
    test_p_4_bb4 = Cpt(EpicsSignalRO, '.BB4')"""


def test_python_gen(test_rec):
    code = '\n'.join(generate_all([test_rec]))
    assert code == _target_code


def test_table_gen(test_rec):

    with NamedTemporaryFile(suffix='.dbd', mode='wt',
                            encoding='utf-8') as f1:
        f1.write('\n'.join(stream_table(test_rec)))
        f1.flush()
        dd = load_dbd_derived(f1.name)
        rec_name = test_rec.name.capitalize() + 'Record'
        code = '\n'.join(generate(rec_name, dd,
                                  super_='RecordBase',
                                  skip_fields=_BASE_FIELDS,
                                  record_type=test_rec.name))
    assert _IMPORTS + '\n' + code == _target_code


_pt_target = """\
+------+------------+----------+
| name | prompt     | dbf_type |
+------+------------+----------+
| AA0  | "test p 0" | DBF_CHAR |
| AA1  | "test p 1" | DBF_CHAR |
| AA2  | "test p 2" | DBF_CHAR |
| AA3  | "test p 3" | DBF_CHAR |
| AA4  | "test p 4" | DBF_CHAR |
| BB0  | "test p 0" | DBF_CHAR |
| BB1  | "test p 1" | DBF_CHAR |
| BB2  | "test p 2" | DBF_CHAR |
| BB3  | "test p 3" | DBF_CHAR |
| BB4  | "test p 4" | DBF_CHAR |
+------+------------+----------+
"""


def test_pretty_table(test_rec):
    pt = prettytable_summary(test_rec, sort=True)
    with TemporaryFile(mode='w+t') as fout:
        print(pt, file=fout)
        fout.seek(0)
        tb = fout.read()

    assert tb == _pt_target
